package org.processmining.localprocessmodelconformance.plugins;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.plugins.IM;

@Plugin(name = "Greedily Explain Log with CloGSgrow patterns", parameterLabels = { "Log" }, returnLabels = { "Local Process Model Collection from CloGSgrow" }, returnTypes = { LocalProcessModelRanking.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class CloGSgrowLogExplainer {
	private Map<XTrace, Set<XTrace>> mergedTracesMap;
	private Map<Integer, Set<Integer>> alreadyExplainedEvents;
	private final double nonOverlappingThreshold = 0.7;

	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Greedily Explain Log with CloGSgrow patterns", requiredParameterLabels = { 0 })
	public LocalProcessModelRanking run(PluginContext context, XLog log) {
		alreadyExplainedEvents = new HashMap<Integer, Set<Integer>>();
		mergedTracesMap = new HashMap<XTrace, Set<XTrace>>();
		
		boolean run = true;
		XLog clone = (XLog) log.clone();
		while(run){
			run = false;
			int highestNumExplainedEvents = 0;
			XTrace bestExplainerOfUnexplainedEvents = null;
			for(int i=0; i<clone.size(); i++){
				int newExplainedEvents = calculateNewExplainedEvents(clone.get(i));
				if(newExplainedEvents>highestNumExplainedEvents){
					highestNumExplainedEvents = newExplainedEvents;
					bestExplainerOfUnexplainedEvents = clone.get(i);
				}
			}
			if(highestNumExplainedEvents>0){
				double lowestNonOverlappingEvents = 1;
				XTrace closestTrace = null;
				for(XTrace existingTrace : mergedTracesMap.keySet()){
					double nonOverlappingEvents = calculcateNonOverlappingEvents(bestExplainerOfUnexplainedEvents, existingTrace);					
					if(nonOverlappingEvents<lowestNonOverlappingEvents){
						lowestNonOverlappingEvents = nonOverlappingEvents;
						closestTrace = existingTrace;
					}
				}
				if(lowestNonOverlappingEvents<nonOverlappingThreshold){
					System.out.println("merging "+getActivityRepresentation(closestTrace)+" and "+getActivityRepresentation(bestExplainerOfUnexplainedEvents)+": "+lowestNonOverlappingEvents);
					clone = mergeEpisodes(new Pair<XTrace,XTrace>(closestTrace, bestExplainerOfUnexplainedEvents), clone);
				}else{
					System.out.println("adding without merge "+getActivityRepresentation(bestExplainerOfUnexplainedEvents));
					clone.remove(bestExplainerOfUnexplainedEvents);
					mergedTracesMap.put(bestExplainerOfUnexplainedEvents, new HashSet<XTrace>());
				}
				updateAlreadyExplainedEvents(bestExplainerOfUnexplainedEvents);
				run = true;
			}
		}
	

		LocalProcessModelRanking ranking = new LocalProcessModelRanking();
		for(XTrace trace : mergedTracesMap.keySet()){
			XLog mergedPatternsLog = newEmptyLog();
			mergedPatternsLog.add(trace);
			Set<XTrace> mergedTraces = mergedTracesMap.get(trace);
			for(XTrace mTrace : mergedTraces)
				mergedPatternsLog.add(mTrace);
			Object[] result = new Object[3];
			if(mergedPatternsLog.size()==1){
				LocalProcessModel lpm = FromPatternSetToLocalProcessModelRanking.traceToNet(trace);
				result[0] = lpm.getAcceptingPetriNet().getNet();
				result[1] = lpm.getAcceptingPetriNet().getInitialMarking();
				result[2] = lpm.getAcceptingPetriNet().getFinalMarkings().iterator().next();
			}else{
				result = IM.minePetriNet(context, mergedPatternsLog, new MiningParametersIM());
			}
			Set<Marking> fMarking = new HashSet<Marking>();
			fMarking.add((Marking) result[2]);
			ranking.addNet(new LocalProcessModel(new AcceptingPetriNetImpl((Petrinet) result[0], (Marking) result[1], fMarking)));
		}
		
		return ranking;
	}

	private void updateAlreadyExplainedEvents(XTrace bestExplainerOfUnexplainedEvents) {
		Map<Integer, Set<Integer>> eventsExplained = getAndParseInstancesAsSet(bestExplainerOfUnexplainedEvents);
		for(Integer key : eventsExplained.keySet()){
			if(alreadyExplainedEvents.containsKey(key)){
				Set<Integer> currentExplained = alreadyExplainedEvents.get(key);
				currentExplained.addAll(eventsExplained.get(key));
				alreadyExplainedEvents.put(key, currentExplained);
			}else{
				alreadyExplainedEvents.put(key, eventsExplained.get(key));
			}			
		}
	}

	private int calculateNewExplainedEvents(XTrace xTrace) {
		Map<Integer, Set<Integer>> t1Map = getAndParseInstancesAsSet(xTrace);
		int count = 0;
		for(Integer key : t1Map.keySet()){
			if(alreadyExplainedEvents.containsKey(key))
				count += CollectionUtils.subtract(t1Map.get(key), alreadyExplainedEvents.get(key)).size();
			else
				count += t1Map.get(key).size();
		}
		return count;
	}

	public static XLog newEmptyLog(){
		return XFactoryRegistry.instance().currentDefault().createLog();
	}
	
	private XLog mergeEpisodes(Pair<XTrace, XTrace> closestTraces, XLog clone) {
		XTrace t1 = closestTraces.getFirst();
		XTrace t2 = closestTraces.getSecond();
		
		Set<XTrace> current = mergedTracesMap.get(t1);
		if(current==null)
			current = new HashSet<XTrace>();
		current.add(t2);
		mergedTracesMap.put(t1, current);
		
		Map<Integer, Set<List<Integer>>> t1Map = getAndParseInstancesAsSetOfLists(t1);
		Map<Integer, Set<List<Integer>>> t2Map = getAndParseInstancesAsSetOfLists(t2);
		
		Set<Integer> keySet = new HashSet<Integer>();
		keySet.addAll(t1Map.keySet());
		keySet.addAll(t2Map.keySet());
		for(Integer key : keySet){
			if(!t1Map.containsKey(key)){
				t1Map.put(key, t2Map.get(key));
				continue;
			}
			if(t2Map.containsKey(key)){
				Set<List<Integer>> t2Candidates = t2Map.get(key);
				for(List<Integer> t2Candidate : t2Candidates){
					boolean inConflict = false;
					int min = t2Candidate.get(0);
					int max = t2Candidate.get(t2Candidate.size()-1);
					for(List<Integer> t1Elements : t1Map.get(key)){
						for(Integer t1Element : t1Elements){
							if(t1Element>min && t1Element<max){
								inConflict = true;
								break;
							}							
						}
						if(inConflict)
							break;
					}
					if(!inConflict)
						t1Map.get(key).add(t2Candidate);
				}
			}
		}
		setInstances(t1, t1Map);
		clone.remove(t2);
		return clone;
	}


	private double calculcateNonOverlappingEvents(XTrace xTrace, XTrace xTrace2) {
		Map<Integer, Set<Integer>> t1Map = getAndParseInstancesAsSet(xTrace);
		Map<Integer, Set<Integer>> t2Map = getAndParseInstancesAsSet(xTrace2);
		return countNonOverlappingEventsFromMaps(t1Map, t2Map);
	}

	private double countNonOverlappingEventsFromMaps(Map<Integer, Set<Integer>> t1Map, Map<Integer, Set<Integer>> t2Map) {
		Set<Integer> keySet = new HashSet<Integer>();
		keySet.addAll(t1Map.keySet());
		keySet.addAll(t2Map.keySet());
		int disjunctionCount = 0;
		for(Integer key : keySet){
			if(!t1Map.keySet().contains(key)){
				disjunctionCount += t2Map.get(key).size();
				continue;
			}
			if(!t2Map.keySet().contains(key)){
				disjunctionCount += t1Map.get(key).size();
				continue;
			}
			disjunctionCount += CollectionUtils.disjunction(t1Map.get(key), t2Map.get(key)).size();
		}
		int unionCount = 0;
		for(Integer key : keySet){
			if(!t1Map.keySet().contains(key)){
				unionCount += t2Map.get(key).size();
				continue;
			}
			if(!t2Map.keySet().contains(key)){
				unionCount += t1Map.get(key).size();
				continue;
			}
			unionCount += CollectionUtils.union(t1Map.get(key), t2Map.get(key)).size();
		}
		return ((double) disjunctionCount)/unionCount;
	}

	private List<String> getActivityRepresentation(XTrace trace) {
		List<String> result = new ArrayList<String>();
		for(XEvent e : trace)
			result.add(getActivity(e));
		return result;
	}
	
	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
	
	private void setInstances(XTrace trace, Map<Integer, Set<List<Integer>>> instances){
		StringBuilder sb = new StringBuilder();
		for(int i : instances.keySet()){
			for(List<Integer> e : instances.get(i)){
				sb.append('('); 
				sb.append(i);
				sb.append(", ");
				sb.append('[');
				boolean first = true;
				for(int el : e){
					if(!first)
						sb.append(',');
					sb.append(el);
					first = false;
				}
				sb.append("]) ");
			}
		}
	}
	
	private Map<Integer, Set<List<Integer>>> getAndParseInstancesAsSetOfLists(XTrace trace){
		Map<Integer, Set<List<Integer>>> explainedEventsPerTraceMap = new HashMap<Integer, Set<List<Integer>>>();
		String instances = getInstances(trace);
		String[] traceIndexEventIndicesPairs = instances.split("]\\)");
		for(String traceIndexEventIndicesPair : traceIndexEventIndicesPairs){
			String[] parts = traceIndexEventIndicesPair.split(",");
			String q = parts[0].replaceAll("\\D+","");
			if(StringUtils.isEmpty(q))
				continue;
			int traceIndex = Integer.parseInt(q);
			Set<List<Integer>> current = explainedEventsPerTraceMap.get(traceIndex);
			if(current==null)
				current = new HashSet<List<Integer>>();
			List<Integer> currentList = new ArrayList<Integer>();
			for(int i=1; i<parts.length; i++)
				currentList.add(Integer.parseInt(parts[i].replaceAll("\\D+","")));
			current.add(currentList);
			explainedEventsPerTraceMap.put(traceIndex, current);
		}
		return explainedEventsPerTraceMap;
	}
	
	private Map<Integer, Set<Integer>> getAndParseInstancesAsSet(XTrace trace){
		Map<Integer, Set<Integer>> explainedEventsPerTraceMap = new HashMap<Integer, Set<Integer>>();
		String instances = getInstances(trace);
		String[] traceIndexEventIndicesPairs = instances.split("]\\)");
		for(String traceIndexEventIndicesPair : traceIndexEventIndicesPairs){
			String[] parts = traceIndexEventIndicesPair.split(",");
			String q = parts[0].replaceAll("\\D+","");
			if(StringUtils.isEmpty(q))
				continue;
			int traceIndex = Integer.parseInt(q);
			Set<Integer> current = explainedEventsPerTraceMap.get(traceIndex);
			if(current==null)
				current = new HashSet<Integer>();
			for(int i=1; i<parts.length; i++)
				current.add(Integer.parseInt(parts[i].replaceAll("\\D+","")));
			explainedEventsPerTraceMap.put(traceIndex, current);
		}
		return explainedEventsPerTraceMap;
	}
	
	private String getInstances(XTrace trace){
		return 	((XAttributeLiteral) trace.getAttributes().get("instances")).getValue();
	}
}