package org.processmining.localprocessmodelconformance.plugins;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.plugins.IM;

@Plugin(name = "CloGSgrow patterns to LPM Collection using Relative least-episode-difference-first merging", parameterLabels = { "Log" }, returnLabels = { "Local Process Model Collection from CloGSgrow" }, returnTypes = { LocalProcessModelRanking.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class LeastRelativeDifferenceFirstCloGSgrowPatternMerging {
	Map<XTrace, Set<XTrace>> mergedTracesMap;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "CloGSgrow patterns to LPM Collection using Relative least-episode-difference-first merging", requiredParameterLabels = { 0 })
	public LocalProcessModelRanking run(PluginContext context, XLog log) {
		mergedTracesMap = new HashMap<XTrace, Set<XTrace>>();
		double nonOverlappingThreshold = 0.30;
		
		boolean run = true;
		XLog clone = (XLog) log.clone();
		while(run){
			double lowestNonOverlappingEvents = 1d;
			Pair<XTrace, XTrace> closestTraces = null;

			run = false;
			int finalI = 0;
			int finalJ = 0;
			for(int i=0; i<clone.size()-1; i++){
				for(int j=i+1; j<clone.size(); j++){
					double nonOverlappingEvents = calculcateNonOverlappingEvents(clone.get(i), clone.get(j));
					if(nonOverlappingEvents<lowestNonOverlappingEvents){
						lowestNonOverlappingEvents = nonOverlappingEvents;
						closestTraces = new Pair<XTrace, XTrace>(clone.get(i), clone.get(j));
						finalI = i;
						finalJ = j;
					}
				}
			}
			if(lowestNonOverlappingEvents<nonOverlappingThreshold){
				System.out.println("merging "+getActivityRepresentation(closestTraces.getFirst())+" and "+getActivityRepresentation(closestTraces.getSecond())+": "+lowestNonOverlappingEvents);
				System.out.println("i: "+finalI+", j: "+finalJ);
				clone = mergeEpisodes(closestTraces, clone);
				run = true;
			}
		}
		
		LocalProcessModelRanking ranking = new LocalProcessModelRanking();
		for(XTrace trace : clone){
			XLog mergedPatternsLog = newEmptyLog();
			mergedPatternsLog.add(trace);
			if(mergedTracesMap.containsKey(trace)){
				Set<XTrace> mergedTraces = mergedTracesMap.get(trace);
				for(XTrace mTrace : mergedTraces)
					mergedPatternsLog.add(mTrace);
			}
			Object[] result = IM.minePetriNet(context, mergedPatternsLog, new MiningParametersIM());
			Set<Marking> fMarking = new HashSet<Marking>();
			fMarking.add((Marking) result[2]);
			ranking.addNet(new LocalProcessModel(new AcceptingPetriNetImpl((Petrinet) result[0], (Marking) result[1], fMarking)));
		}
		
		return ranking;
	}

	public static XLog newEmptyLog(){
		return XFactoryRegistry.instance().currentDefault().createLog();
	}
	
	private XLog mergeEpisodes(Pair<XTrace, XTrace> closestTraces, XLog clone) {
		XTrace t1 = closestTraces.getFirst();
		XTrace t2 = closestTraces.getSecond();
		
		Set<XTrace> current = mergedTracesMap.get(t1);
		if(current==null)
			current = new HashSet<XTrace>();
		current.add(t2);
		mergedTracesMap.put(t1, current);
		
		Map<Integer, Set<List<Integer>>> t1Map = getAndParseInstancesAsSetOfLists(t1);
		Map<Integer, Set<List<Integer>>> t2Map = getAndParseInstancesAsSetOfLists(t2);
		
		Set<Integer> keySet = new HashSet<Integer>();
		keySet.addAll(t1Map.keySet());
		keySet.addAll(t2Map.keySet());
		for(Integer key : keySet){
			if(!t1Map.containsKey(key)){
				t1Map.put(key, t2Map.get(key));
				continue;
			}
			if(t2Map.containsKey(key)){
				Set<List<Integer>> t2Candidates = t2Map.get(key);
				for(List<Integer> t2Candidate : t2Candidates){
					boolean inConflict = false;
					int min = t2Candidate.get(0);
					int max = t2Candidate.get(t2Candidate.size()-1);
					for(List<Integer> t1Elements : t1Map.get(key)){
						for(Integer t1Element : t1Elements){
							if(t1Element>min && t1Element<max){
								inConflict = true;
								break;
							}							
						}
						if(inConflict)
							break;
					}
					if(!inConflict)
						t1Map.get(key).add(t2Candidate);
				}
			}
		}
		setInstances(t1, t1Map);
		clone.remove(t2);
		return clone;
	}


	private double calculcateNonOverlappingEvents(XTrace xTrace, XTrace xTrace2) {
		Map<Integer, Set<Integer>> t1Map = getAndParseInstancesAsSet(xTrace);
		Map<Integer, Set<Integer>> t2Map = getAndParseInstancesAsSet(xTrace2);
		return countNonOverlappingEventsFromMaps(t1Map, t2Map);
	}

	private double countNonOverlappingEventsFromMaps(Map<Integer, Set<Integer>> t1Map, Map<Integer, Set<Integer>> t2Map) {
		Set<Integer> keySet = new HashSet<Integer>();
		keySet.addAll(t1Map.keySet());
		keySet.addAll(t2Map.keySet());
		int disjunctionCount = 0;
		for(Integer key : keySet){
			if(!t1Map.keySet().contains(key)){
				disjunctionCount += t2Map.get(key).size();
				continue;
			}
			if(!t2Map.keySet().contains(key)){
				disjunctionCount += t1Map.get(key).size();
				continue;
			}
			disjunctionCount += CollectionUtils.disjunction(t1Map.get(key), t2Map.get(key)).size();
		}
		int unionCount = 0;
		for(Integer key : keySet){
			if(!t1Map.keySet().contains(key)){
				unionCount += t2Map.get(key).size();
				continue;
			}
			if(!t2Map.keySet().contains(key)){
				unionCount += t1Map.get(key).size();
				continue;
			}
			unionCount += CollectionUtils.union(t1Map.get(key), t2Map.get(key)).size();
		}
		return ((double) disjunctionCount)/unionCount;
	}

	private List<String> getActivityRepresentation(XTrace trace) {
		List<String> result = new ArrayList<String>();
		for(XEvent e : trace)
			result.add(getActivity(e));
		return result;
	}
	
	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
	
	private void setInstances(XTrace trace, Map<Integer, Set<List<Integer>>> instances){
		StringBuilder sb = new StringBuilder();
		for(int i : instances.keySet()){
			for(List<Integer> e : instances.get(i)){
				sb.append('('); 
				sb.append(i);
				sb.append(", ");
				sb.append('[');
				boolean first = true;
				for(int el : e){
					if(!first)
						sb.append(',');
					sb.append(el);
					first = false;
				}
				sb.append("]) ");
			}
		}
	}
	
	private Map<Integer, Set<List<Integer>>> getAndParseInstancesAsSetOfLists(XTrace trace){
		Map<Integer, Set<List<Integer>>> explainedEventsPerTraceMap = new HashMap<Integer, Set<List<Integer>>>();
		String instances = getInstances(trace);
		String[] traceIndexEventIndicesPairs = instances.split("]\\)");
		for(String traceIndexEventIndicesPair : traceIndexEventIndicesPairs){
			String[] parts = traceIndexEventIndicesPair.split(",");
			String q = parts[0].replaceAll("\\D+","");
			if(StringUtils.isEmpty(q))
				continue;
			int traceIndex = Integer.parseInt(q);
			Set<List<Integer>> current = explainedEventsPerTraceMap.get(traceIndex);
			if(current==null)
				current = new HashSet<List<Integer>>();
			List<Integer> currentList = new ArrayList<Integer>();
			for(int i=1; i<parts.length; i++)
				currentList.add(Integer.parseInt(parts[i].replaceAll("\\D+","")));
			current.add(currentList);
			explainedEventsPerTraceMap.put(traceIndex, current);
		}
		return explainedEventsPerTraceMap;
	}
	
	private Map<Integer, Set<Integer>> getAndParseInstancesAsSet(XTrace trace){
		Map<Integer, Set<Integer>> explainedEventsPerTraceMap = new HashMap<Integer, Set<Integer>>();
		String instances = getInstances(trace);
		String[] traceIndexEventIndicesPairs = instances.split("]\\)");
		for(String traceIndexEventIndicesPair : traceIndexEventIndicesPairs){
			String[] parts = traceIndexEventIndicesPair.split(",");
			String q = parts[0].replaceAll("\\D+","");
			if(StringUtils.isEmpty(q))
				continue;
			int traceIndex = Integer.parseInt(q);
			Set<Integer> current = explainedEventsPerTraceMap.get(traceIndex);
			if(current==null)
				current = new HashSet<Integer>();
			for(int i=1; i<parts.length; i++)
				current.add(Integer.parseInt(parts[i].replaceAll("\\D+","")));
			explainedEventsPerTraceMap.put(traceIndex, current);
		}
		return explainedEventsPerTraceMap;
	}
	
	private String getInstances(XTrace trace){
		return 	((XAttributeLiteral) trace.getAttributes().get("instances")).getValue();
	}
}
