package org.processmining.localprocessmodelconformance.plugins;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeDiscreteImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;

@Plugin(name = "Merge Parallel", parameterLabels = { "Log" }, returnLabels = { "Log with Parallel Merged" }, returnTypes = { XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class MergeParallel {
	Map<XTrace, Set<XTrace>> identicalUnderParallelMap;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Merge Parallel", requiredParameterLabels = { 0 })
	public XLog run(PluginContext context, XLog log) {
		XLog clone = (XLog) log.clone();
		identicalUnderParallelMap = new HashMap<XTrace, Set<XTrace>>();
		
		for(int i=0; i<clone.size(); i++){
			XTrace ti = clone.get(i);
			for(int j=0; j<clone.size(); j++){
				XTrace tj = clone.get(j);
				int minSize = Math.min(ti.size(), tj.size());
				for(int n=1; n<minSize/2; n++){
					// see if identical when we allow a block of at most size n to be unordered.
					if(isParMatchForSizeN(ti, tj, n)){
						addToIdenticalMap(ti, tj);
						addToIdenticalMap(tj, ti);
					}
				}
			}
		}
		
		collapseIdenticalTraces(clone);
		return clone;
	}

	private void addToIdenticalMap(XTrace ti, XTrace tj) {
		Set<XTrace> current = identicalUnderParallelMap.get(ti);
		if(current==null)
			current = new HashSet<XTrace>();
		current.add(tj);
		identicalUnderParallelMap.put(ti, current);
	}
	
	private boolean isParMatchForSizeN(XTrace ti, XTrace tj, int n) {
		if(ti.size()!=tj.size())
			return false;
		List<String> actsI = getActivityRepresentation(ti);
		List<String> actsJ = getActivityRepresentation(tj);
		for(int i=0; i< actsI.size()-n; i++){
			if(getParAbstractionOfSizeNAtIndexI(actsI, n, i).equals(getParAbstractionOfSizeNAtIndexI(actsJ, n, i)))
				return true;
		}
		return false;
	}

	private List<Set<String>> getParAbstractionOfSizeNAtIndexI(List<String> actsJ, int n, int i) {
		int j=0;
		List<Set<String>> abstraction = new ArrayList<Set<String>>();
		while(j<actsJ.size()){
			if(j!=i){
				Set<String> singleton = new HashSet<String>();
				singleton.add(actsJ.get(j));
				abstraction.add(singleton);
				j++;
			}else{
				int k=0;
				Set<String> parBlock = new HashSet<String>();
				while(k<n){
					parBlock.add(actsJ.get(j));
					k++; j++;
				}
				abstraction.add(parBlock);
			}
		}
		return abstraction;
	}

	private void collapseIdenticalTraces(XLog log){
		Set<XTrace> toRemoveSet = new HashSet<XTrace>();
		for(int i=0; i<log.size(); i++){
			for(int j=(i+1); j<log.size(); j++){
				if(getActivityRepresentation(log.get(i)).equals(getActivityRepresentation(log.get(j)))){
					if(getMerged(log.get(j))){
						toRemoveSet.add(log.get(i));
						setSupport(log.get(j), getSupport(log.get(i))+getSupport(log.get(j)));
					}else{
						toRemoveSet.add(log.get(j));
						setSupport(log.get(i), getSupport(log.get(i))+getSupport(log.get(j)));
					}
				}
			}
		}
		for(XTrace t : toRemoveSet)
			log.remove(t);
		toRemoveSet.clear();
		for(XTrace t: log)
			if(t.size()==1)
				toRemoveSet.add(t);
		for(XTrace t : toRemoveSet)
			log.remove(t);
	}
	
	private List<String> getActivityRepresentation(XTrace trace) {
		List<String> result = new ArrayList<String>();
		for(XEvent e : trace)
			result.add(getActivity(e));
		return result;
	}

	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
	
	public static boolean getMerged(XTrace trace){
		return trace.getAttributes().containsKey("merged");
	}
	
	public static void setMerged(XTrace trace){
		trace.getAttributes().put("merged", new XAttributeLiteralImpl("merged", "TRUE"));
	}
	
	public static void setSupport(XTrace trace, int value){
		trace.getAttributes().put("support", new XAttributeDiscreteImpl("support", value));
	}
	
	public static int getSupport(XTrace trace){
		return (int) ((XAttributeDiscrete) trace.getAttributes().get("support")).getValue();
	}
	
}
